from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class AutomationEnvelope:
    """
    Describes parameter automation per clip.
    """

    def insert_step(self, arg2: float, arg3: float, arg4: float) -> None:
        """
        insert_step( (AutomationEnvelope)arg1, (float)arg2, (float)arg3, (float)arg4) -> None :

            C++ signature :
                void insert_step(AAutomation {lvalue},double,double,double)
        """

    def value_at_time(self, arg2: float) -> float:
        """
        value_at_time( (AutomationEnvelope)arg1, (float)arg2) -> float :

            C++ signature :
                double value_at_time(AAutomation {lvalue},double)
        """


class Clip:
    """
    This class represents a Clip in Live. It can be either an Audio
    Clip or a MIDI Clip, in an Arrangement or the Session, depending
    on the Track (Slot) it lives in.
    """

    class View:
        """
        Representing the view aspects of a Clip.
        """

        def hide_envelope(self) -> None:
            """
            hide_envelope( (View)arg1) -> None :
                Hide the envelope view.

                C++ signature :
                    void hide_envelope(TPyViewData<AClip>)
            """

        def select_envelope_parameter(self, arg2: object) -> None:
            """
            select_envelope_parameter( (View)arg1, (DeviceParameter)arg2) -> None :
                Select the given device parameter in the envelope view.

                C++ signature :
                    void select_envelope_parameter(TPyViewData<AClip>,TPyHandle<ATimeableValue>)
            """

        def show_envelope(self) -> None:
            """
            show_envelope( (View)arg1) -> None :
                Show the envelope view.

                C++ signature :
                    void show_envelope(TPyViewData<AClip>)
            """

        def show_loop(self) -> None:
            """
            show_loop( (View)arg1) -> None :
                Show the entire loop in the detail view.

                C++ signature :
                    void show_loop(TPyViewData<AClip>)
            """

        @property
        def canonical_parent(self) -> Any:
            """
            Get the canonical parent of the clip view.
            """

        @property
        def grid_is_triplet(self) -> Any:
            """
            Get/set wether the grid is showing in triplet mode.
            """

        @grid_is_triplet.setter
        def grid_is_triplet(self, value: Any) -> None:
            pass

        @property
        def grid_quantization(self) -> Any:
            """
            Get/set clip grid quantization resolution.
            """

        @grid_quantization.setter
        def grid_quantization(self, value: Any) -> None:
            pass

    def add_color_index_listener(self, listener: Callable) -> None:
        """
        add_color_index_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "color_index" has changed.

            C++ signature :
                void add_color_index_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_color_listener(self, listener: Callable) -> None:
        """
        add_color_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "color" has changed.

            C++ signature :
                void add_color_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_end_marker_listener(self, listener: Callable) -> None:
        """
        add_end_marker_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "end_marker" has changed.

            C++ signature :
                void add_end_marker_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_end_time_listener(self, listener: Callable) -> None:
        """
        add_end_time_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "end_time" has changed.

            C++ signature :
                void add_end_time_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_file_path_listener(self, listener: Callable) -> None:
        """
        add_file_path_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "file_path" has changed.

            C++ signature :
                void add_file_path_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_gain_listener(self, listener: Callable) -> None:
        """
        add_gain_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "gain" has changed.

            C++ signature :
                void add_gain_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_groove_listener(self, listener: Callable) -> None:
        """
        add_groove_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "groove" has changed.

            C++ signature :
                void add_groove_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_has_envelopes_listener(self, listener: Callable) -> None:
        """
        add_has_envelopes_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "has_envelopes" has changed.

            C++ signature :
                void add_has_envelopes_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_is_overdubbing_listener(self, listener: Callable) -> None:
        """
        add_is_overdubbing_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_overdubbing" has changed.

            C++ signature :
                void add_is_overdubbing_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_is_recording_listener(self, listener: Callable) -> None:
        """
        add_is_recording_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_recording" has changed.

            C++ signature :
                void add_is_recording_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_launch_mode_listener(self, listener: Callable) -> None:
        """
        add_launch_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "launch_mode" has changed.

            C++ signature :
                void add_launch_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_launch_quantization_listener(self, listener: Callable) -> None:
        """
        add_launch_quantization_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "launch_quantization" has changed.

            C++ signature :
                void add_launch_quantization_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_legato_listener(self, listener: Callable) -> None:
        """
        add_legato_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "legato" has changed.

            C++ signature :
                void add_legato_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_loop_end_listener(self, listener: Callable) -> None:
        """
        add_loop_end_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "loop_end" has changed.

            C++ signature :
                void add_loop_end_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_loop_jump_listener(self, listener: Callable) -> None:
        """
        add_loop_jump_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "loop_jump" has changed.

            C++ signature :
                void add_loop_jump_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_loop_start_listener(self, listener: Callable) -> None:
        """
        add_loop_start_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "loop_start" has changed.

            C++ signature :
                void add_loop_start_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_looping_listener(self, listener: Callable) -> None:
        """
        add_looping_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "looping" has changed.

            C++ signature :
                void add_looping_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_muted_listener(self, listener: Callable) -> None:
        """
        add_muted_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "muted" has changed.

            C++ signature :
                void add_muted_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_new_notes(self, arg2: object) -> IntU64Vector:
        """
        add_new_notes( (Clip)arg1, (object)arg2) -> IntU64Vector :
            Expects a Python iterable holding a number of Live.Clip.MidiNoteSpecification
            objects. The objects will be used to construct new notes in the clip.

            C++ signature :
                std::__1::vector<unsigned long long, std::__1::allocator<unsigned long long>> add_new_notes(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_notes_listener(self, listener: Callable) -> None:
        """
        add_notes_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "notes" has changed.

            C++ signature :
                void add_notes_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_pitch_coarse_listener(self, listener: Callable) -> None:
        """
        add_pitch_coarse_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "pitch_coarse" has changed.

            C++ signature :
                void add_pitch_coarse_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_pitch_fine_listener(self, listener: Callable) -> None:
        """
        add_pitch_fine_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "pitch_fine" has changed.

            C++ signature :
                void add_pitch_fine_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_playing_position_listener(self, listener: Callable) -> None:
        """
        add_playing_position_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "playing_position" has changed.

            C++ signature :
                void add_playing_position_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_playing_status_listener(self, listener: Callable) -> None:
        """
        add_playing_status_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "playing_status" has changed.

            C++ signature :
                void add_playing_status_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_position_listener(self, listener: Callable) -> None:
        """
        add_position_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "position" has changed.

            C++ signature :
                void add_position_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_ram_mode_listener(self, listener: Callable) -> None:
        """
        add_ram_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "ram_mode" has changed.

            C++ signature :
                void add_ram_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_signature_denominator_listener(self, listener: Callable) -> None:
        """
        add_signature_denominator_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "signature_denominator" has changed.

            C++ signature :
                void add_signature_denominator_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_signature_numerator_listener(self, listener: Callable) -> None:
        """
        add_signature_numerator_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "signature_numerator" has changed.

            C++ signature :
                void add_signature_numerator_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_start_marker_listener(self, listener: Callable) -> None:
        """
        add_start_marker_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "start_marker" has changed.

            C++ signature :
                void add_start_marker_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_velocity_amount_listener(self, listener: Callable) -> None:
        """
        add_velocity_amount_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "velocity_amount" has changed.

            C++ signature :
                void add_velocity_amount_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_warp_marker(self, warp_marker: object) -> None:
        """
        add_warp_marker( (Clip)self, (object)warp_marker) -> None :
            Available for AudioClips only.
            Adds the specified warp marker, if possible.

            C++ signature :
                void add_warp_marker(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_warp_markers_listener(self, listener: Callable) -> None:
        """
        add_warp_markers_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warp_markers" has changed.

            C++ signature :
                void add_warp_markers_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_warp_mode_listener(self, listener: Callable) -> None:
        """
        add_warp_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warp_mode" has changed.

            C++ signature :
                void add_warp_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def add_warping_listener(self, listener: Callable) -> None:
        """
        add_warping_listener( (Clip)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warping" has changed.

            C++ signature :
                void add_warping_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def apply_note_modifications(self, arg2: object) -> None:
        """
        apply_note_modifications( (Clip)arg1, (MidiNoteVector)arg2) -> None :
            Expects a list of notes as returned from get_notes_extended. The content
            of the list will be used to modify existing notes in the clip, based on
            matching note IDs.
            This function should be used when modifying existing notes, e.g. changing the
            velocity or start time. The function ensures that per-note events attached to
            the modified notes are preserved. This is NOT the case when replacing notes
            via a combination of remove_notes_extended and add_new_notes.
            The given list can be a subset of the notes in the clip, but it must not
            contain any notes that are not present in the clip.


            C++ signature :
                void apply_note_modifications(TPyHandle<AClip>,std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>>)
        """

    def automation_envelope(self, arg2: object) -> AutomationEnvelope:
        """
        automation_envelope( (Clip)arg1, (DeviceParameter)arg2) -> AutomationEnvelope :
            Return the envelope for the given parameter.Returns None if the envelope doesn't exist.Returns None for Arrangement clips.Returns None for parameters from a different track.

            C++ signature :
                TWeakPtr<AAutomation> automation_envelope(TPyHandle<AClip>,TPyHandle<ATimeableValue>)
        """

    def beat_to_sample_time(self, beat_time: float) -> float:
        """
        beat_to_sample_time( (Clip)self, (float)beat_time) -> float :
            Available for AudioClips only.
            Converts the given beat time to sample time. Raises an error if the sample is not warped.

            C++ signature :
                double beat_to_sample_time(TPyHandle<AClip>,double)
        """

    def clear_all_envelopes(self) -> None:
        """
        clear_all_envelopes( (Clip)arg1) -> None :
            Clears all envelopes for this clip.

            C++ signature :
                void clear_all_envelopes(TPyHandle<AClip>)
        """

    def clear_envelope(self, arg2: object) -> None:
        """
        clear_envelope( (Clip)arg1, (DeviceParameter)arg2) -> None :
            Clears the envelope of this clips given parameter.

            C++ signature :
                void clear_envelope(TPyHandle<AClip>,TPyHandle<ATimeableValue>)
        """

    def color_has_listener(self, listener: Callable) -> bool:
        """
        color_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "color".

            C++ signature :
                bool color_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def color_index_has_listener(self, listener: Callable) -> bool:
        """
        color_index_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "color_index".

            C++ signature :
                bool color_index_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def create_automation_envelope(self, arg2: object) -> AutomationEnvelope:
        """
        create_automation_envelope( (Clip)arg1, (DeviceParameter)arg2) -> AutomationEnvelope :
            Creates an envelope for a given parameter and returns it.This should only be used if the envelope doesn't exist.Raises an error if the envelope can't be created.

            C++ signature :
                TWeakPtr<AAutomation> create_automation_envelope(TPyHandle<AClip>,TPyHandle<ATimeableValue>)
        """

    def crop(self) -> None:
        """
        crop( (Clip)arg1) -> None :
            Crops the clip. The region that is cropped depends on whether the clip is
            looped or not. If looped, the region outside of the loop is removed.
            If not looped, the region outside the start and end markers is removed.

            C++ signature :
                void crop(TPyHandle<AClip>)
        """

    def deselect_all_notes(self) -> None:
        """
        deselect_all_notes( (Clip)arg1) -> None :
            De-selects all notes present in the clip.

            C++ signature :
                void deselect_all_notes(TPyHandle<AClip>)
        """

    def duplicate_loop(self) -> None:
        """
        duplicate_loop( (Clip)arg1) -> None :
            Make the loop two times longer and duplicates notes and envelopes.
            Duplicates the clip start/end range if the clip is not looped.

            C++ signature :
                void duplicate_loop(TPyHandle<AClip>)
        """

    def duplicate_notes_by_id(
        self, note_ids: object, destination_time: object = None, transposition_amount: int = 0
    ) -> IntU64Vector:
        """
        duplicate_notes_by_id( (Clip)self, (object)note_ids [, (object)destination_time=None [, (int)transposition_amount=0]]) -> IntU64Vector :
            Duplicate all notes matching the given note IDs.
            If the optional destination_time is not provided, new notes will be inserted
            after the last selected note. This behavior can be observed when duplicating
            notes in the Live GUI.
            If the transposition_amount is specified, the notes in the region will be
            transposed by the number of semitones.
            Raises an error on audio clips.

            C++ signature :
                std::__1::vector<unsigned long long, std::__1::allocator<unsigned long long>> duplicate_notes_by_id(TPyHandle<AClip>,boost::python::api::object [,boost::python::api::object=None [,int=0]])
        """

    def duplicate_region(
        self,
        region_start: float,
        region_length: float,
        destination_time: float,
        pitch: int = 0,
        transposition_amount: int = 0,
    ) -> None:
        """
        duplicate_region( (Clip)self, (float)region_start, (float)region_length, (float)destination_time [, (int)pitch=-1 [, (int)transposition_amount=0]]) -> None :
            Duplicate the notes in the specified region to the destination_time.
            Only notes of the specified pitch are duplicated or all if pitch is -1.
            If the transposition_amount is not 0, the notes in the region will
            be transposed by the transpose_amount of semitones.Raises an error on audio clips.

            C++ signature :
                void duplicate_region(TPyHandle<AClip>,double,double,double [,int=-1 [,int=0]])
        """

    def end_marker_has_listener(self, listener: Callable) -> bool:
        """
        end_marker_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "end_marker".

            C++ signature :
                bool end_marker_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def end_time_has_listener(self, listener: Callable) -> bool:
        """
        end_time_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "end_time".

            C++ signature :
                bool end_time_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def file_path_has_listener(self, listener: Callable) -> bool:
        """
        file_path_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "file_path".

            C++ signature :
                bool file_path_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def fire(self) -> None:
        """
        fire( (Clip)arg1) -> None :
            (Re)Start playing this Clip.

            C++ signature :
                void fire(TPyHandle<AClip>)
        """

    def gain_has_listener(self, listener: Callable) -> bool:
        """
        gain_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "gain".

            C++ signature :
                bool gain_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def get_all_notes_extended(self) -> MidiNoteVector:
        """
        get_all_notes_extended( (Clip)arg1) -> MidiNoteVector :
            Returns a list of all MIDI notes from the clip, regardless of their position
            relative to the start and end markers/loop start and loop end.
            Each note is represented by a Live.Clip.MidiNote object.
            The returned list can be modified freely, but modifications will not
            be reflected in the MIDI clip until apply_note_modifications is called.

            C++ signature :
                std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> get_all_notes_extended(TPyHandle<AClip>)
        """

    def get_notes(self, from_time: float, from_pitch: int, time_span: float, pitch_span: int) -> tuple:
        """
        get_notes( (Clip)self, (float)from_time, (int)from_pitch, (float)time_span, (int)pitch_span) -> tuple :
            Returns a tuple of tuples where each inner tuple represents
            a note starting in the given pitch- and time range.
            The inner tuple contains pitch, time, duration, velocity, and mute state.

            C++ signature :
                boost::python::tuple get_notes(TPyHandle<AClip>,double,int,double,int)
        """

    def get_notes_by_id(self, note_ids: object) -> MidiNoteVector:
        """
        get_notes_by_id( (Clip)arg1, (object)note_ids) -> MidiNoteVector :
            Return a list of MIDI notes matching the given note IDs.


            C++ signature :
                std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> get_notes_by_id(TPyHandle<AClip>,boost::python::api::object)
        """

    def get_notes_extended(
        self, from_pitch: int, pitch_span: int, from_time: float, time_span: float
    ) -> MidiNoteVector:
        """
        get_notes_extended( (Clip)arg1, (int)from_pitch, (int)pitch_span, (float)from_time, (float)time_span) -> MidiNoteVector :
            Returns a list of MIDI notes from the given pitch and time range.
            Each note is represented by a Live.Clip.MidiNote object.
            The returned list can be modified freely, but modifications will not
            be reflected in the MIDI clip until apply_note_modifications is called.

            C++ signature :
                std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> get_notes_extended(TPyHandle<AClip>,int,int,double,double)
        """

    def get_selected_notes(self) -> tuple:
        """
        get_selected_notes( (Clip)arg1) -> tuple :
            Returns a tuple of tuples where each inner tuple
            represents a selected note. The inner tuple contains
            pitch, time, duration, velocity, and mute state.

            C++ signature :
                boost::python::tuple get_selected_notes(TPyHandle<AClip>)
        """

    def get_selected_notes_extended(self) -> MidiNoteVector:
        """
        get_selected_notes_extended( (Clip)arg1) -> MidiNoteVector :
            Returns a list of all MIDI notes from the clip that are currently selected.
            Each note is represented by a Live.Clip.MidiNote object.
            The returned list can be modified freely, but modifications will not
            be reflected in the MIDI clip until apply_note_modifications is called.

            C++ signature :
                std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> get_selected_notes_extended(TPyHandle<AClip>)
        """

    def groove_has_listener(self, listener: Callable) -> bool:
        """
        groove_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "groove".

            C++ signature :
                bool groove_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def has_envelopes_has_listener(self, listener: Callable) -> bool:
        """
        has_envelopes_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "has_envelopes".

            C++ signature :
                bool has_envelopes_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def is_overdubbing_has_listener(self, listener: Callable) -> bool:
        """
        is_overdubbing_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_overdubbing".

            C++ signature :
                bool is_overdubbing_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def is_recording_has_listener(self, listener: Callable) -> bool:
        """
        is_recording_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_recording".

            C++ signature :
                bool is_recording_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def launch_mode_has_listener(self, listener: Callable) -> bool:
        """
        launch_mode_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "launch_mode".

            C++ signature :
                bool launch_mode_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def launch_quantization_has_listener(self, listener: Callable) -> bool:
        """
        launch_quantization_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "launch_quantization".

            C++ signature :
                bool launch_quantization_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def legato_has_listener(self, listener: Callable) -> bool:
        """
        legato_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "legato".

            C++ signature :
                bool legato_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def loop_end_has_listener(self, listener: Callable) -> bool:
        """
        loop_end_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "loop_end".

            C++ signature :
                bool loop_end_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def loop_jump_has_listener(self, listener: Callable) -> bool:
        """
        loop_jump_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "loop_jump".

            C++ signature :
                bool loop_jump_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def loop_start_has_listener(self, listener: Callable) -> bool:
        """
        loop_start_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "loop_start".

            C++ signature :
                bool loop_start_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def looping_has_listener(self, listener: Callable) -> bool:
        """
        looping_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "looping".

            C++ signature :
                bool looping_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def move_playing_pos(self, arg2: float) -> None:
        """
        move_playing_pos( (Clip)arg1, (float)arg2) -> None :
            Jump forward or backward by the specified relative amount in beats.
            Will do nothing, if the Clip is not playing.

            C++ signature :
                void move_playing_pos(TPyHandle<AClip>,double)
        """

    def move_warp_marker(self, marker_beat_time: float, beat_time_distance: float) -> None:
        """
        move_warp_marker( (Clip)self, (float)marker_beat_time, (float)beat_time_distance) -> None :
            Available for AudioClips only.
            Moves the specified warp marker by the specified beat time amount, if possible.

            C++ signature :
                void move_warp_marker(TPyHandle<AClip>,double,double)
        """

    def muted_has_listener(self, listener: Callable) -> bool:
        """
        muted_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "muted".

            C++ signature :
                bool muted_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def notes_has_listener(self, listener: Callable) -> bool:
        """
        notes_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "notes".

            C++ signature :
                bool notes_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def pitch_coarse_has_listener(self, listener: Callable) -> bool:
        """
        pitch_coarse_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "pitch_coarse".

            C++ signature :
                bool pitch_coarse_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def pitch_fine_has_listener(self, listener: Callable) -> bool:
        """
        pitch_fine_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "pitch_fine".

            C++ signature :
                bool pitch_fine_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def playing_position_has_listener(self, listener: Callable) -> bool:
        """
        playing_position_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "playing_position".

            C++ signature :
                bool playing_position_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def playing_status_has_listener(self, listener: Callable) -> bool:
        """
        playing_status_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "playing_status".

            C++ signature :
                bool playing_status_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def position_has_listener(self, listener: Callable) -> bool:
        """
        position_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "position".

            C++ signature :
                bool position_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def quantize(self, arg2: int, arg3: float) -> None:
        """
        quantize( (Clip)arg1, (int)arg2, (float)arg3) -> None :
            Quantize all notes in a clip or align warp markers.

            C++ signature :
                void quantize(TPyHandle<AClip>,int,float)
        """

    def quantize_pitch(self, arg2: int, arg3: int, arg4: float) -> None:
        """
        quantize_pitch( (Clip)arg1, (int)arg2, (int)arg3, (float)arg4) -> None :
            Quantize all the notes of a given pitch.  Raises an error on audio clips.

            C++ signature :
                void quantize_pitch(TPyHandle<AClip>,int,int,float)
        """

    def ram_mode_has_listener(self, listener: Callable) -> bool:
        """
        ram_mode_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "ram_mode".

            C++ signature :
                bool ram_mode_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_color_index_listener(self, listener: Callable) -> None:
        """
        remove_color_index_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "color_index".

            C++ signature :
                void remove_color_index_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_color_listener(self, listener: Callable) -> None:
        """
        remove_color_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "color".

            C++ signature :
                void remove_color_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_end_marker_listener(self, listener: Callable) -> None:
        """
        remove_end_marker_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "end_marker".

            C++ signature :
                void remove_end_marker_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_end_time_listener(self, listener: Callable) -> None:
        """
        remove_end_time_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "end_time".

            C++ signature :
                void remove_end_time_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_file_path_listener(self, listener: Callable) -> None:
        """
        remove_file_path_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "file_path".

            C++ signature :
                void remove_file_path_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_gain_listener(self, listener: Callable) -> None:
        """
        remove_gain_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "gain".

            C++ signature :
                void remove_gain_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_groove_listener(self, listener: Callable) -> None:
        """
        remove_groove_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "groove".

            C++ signature :
                void remove_groove_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_has_envelopes_listener(self, listener: Callable) -> None:
        """
        remove_has_envelopes_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "has_envelopes".

            C++ signature :
                void remove_has_envelopes_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_is_overdubbing_listener(self, listener: Callable) -> None:
        """
        remove_is_overdubbing_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_overdubbing".

            C++ signature :
                void remove_is_overdubbing_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_is_recording_listener(self, listener: Callable) -> None:
        """
        remove_is_recording_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_recording".

            C++ signature :
                void remove_is_recording_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_launch_mode_listener(self, listener: Callable) -> None:
        """
        remove_launch_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "launch_mode".

            C++ signature :
                void remove_launch_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_launch_quantization_listener(self, listener: Callable) -> None:
        """
        remove_launch_quantization_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "launch_quantization".

            C++ signature :
                void remove_launch_quantization_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_legato_listener(self, listener: Callable) -> None:
        """
        remove_legato_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "legato".

            C++ signature :
                void remove_legato_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_loop_end_listener(self, listener: Callable) -> None:
        """
        remove_loop_end_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "loop_end".

            C++ signature :
                void remove_loop_end_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_loop_jump_listener(self, listener: Callable) -> None:
        """
        remove_loop_jump_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "loop_jump".

            C++ signature :
                void remove_loop_jump_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_loop_start_listener(self, listener: Callable) -> None:
        """
        remove_loop_start_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "loop_start".

            C++ signature :
                void remove_loop_start_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_looping_listener(self, listener: Callable) -> None:
        """
        remove_looping_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "looping".

            C++ signature :
                void remove_looping_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_muted_listener(self, listener: Callable) -> None:
        """
        remove_muted_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "muted".

            C++ signature :
                void remove_muted_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_notes(self, arg2: float, arg3: int, arg4: float, arg5: int) -> None:
        """
        remove_notes( (Clip)arg1, (float)arg2, (int)arg3, (float)arg4, (int)arg5) -> None :
            Delete all notes starting in the given pitch- and time range.

            C++ signature :
                void remove_notes(TPyHandle<AClip>,double,int,double,int)
        """

    def remove_notes_by_id(self, arg2: object) -> None:
        """
        remove_notes_by_id( (Clip)arg1, (object)arg2) -> None :
            Delete all notes matching the given note IDs.
            This function should NOT be used to implement modification of existing notes
            (i.e. in combination with add_new_notes), as that leads to loss of per-note
            events. apply_note_modifications must be used instead for modifying existing
            notes.

            C++ signature :
                void remove_notes_by_id(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_notes_extended(self, from_pitch: int, pitch_span: int, from_time: float, time_span: float) -> None:
        """
        remove_notes_extended( (Clip)arg1, (int)from_pitch, (int)pitch_span, (float)from_time, (float)time_span) -> None :
            Delete all notes starting in the given pitch and time range.
            This function should NOT be used to implement modification of existing notes
            (i.e. in combination with add_new_notes), as that leads to loss of per-note
            events. apply_note_modifications must be used instead for modifying existing
            notes.

            C++ signature :
                void remove_notes_extended(TPyHandle<AClip>,int,int,double,double)
        """

    def remove_notes_listener(self, listener: Callable) -> None:
        """
        remove_notes_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "notes".

            C++ signature :
                void remove_notes_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_pitch_coarse_listener(self, listener: Callable) -> None:
        """
        remove_pitch_coarse_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "pitch_coarse".

            C++ signature :
                void remove_pitch_coarse_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_pitch_fine_listener(self, listener: Callable) -> None:
        """
        remove_pitch_fine_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "pitch_fine".

            C++ signature :
                void remove_pitch_fine_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_playing_position_listener(self, listener: Callable) -> None:
        """
        remove_playing_position_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "playing_position".

            C++ signature :
                void remove_playing_position_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_playing_status_listener(self, listener: Callable) -> None:
        """
        remove_playing_status_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "playing_status".

            C++ signature :
                void remove_playing_status_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_position_listener(self, listener: Callable) -> None:
        """
        remove_position_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "position".

            C++ signature :
                void remove_position_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_ram_mode_listener(self, listener: Callable) -> None:
        """
        remove_ram_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "ram_mode".

            C++ signature :
                void remove_ram_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_signature_denominator_listener(self, listener: Callable) -> None:
        """
        remove_signature_denominator_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "signature_denominator".

            C++ signature :
                void remove_signature_denominator_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_signature_numerator_listener(self, listener: Callable) -> None:
        """
        remove_signature_numerator_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "signature_numerator".

            C++ signature :
                void remove_signature_numerator_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_start_marker_listener(self, listener: Callable) -> None:
        """
        remove_start_marker_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "start_marker".

            C++ signature :
                void remove_start_marker_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_velocity_amount_listener(self, listener: Callable) -> None:
        """
        remove_velocity_amount_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "velocity_amount".

            C++ signature :
                void remove_velocity_amount_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_warp_marker(self, beat_time: float) -> None:
        """
        remove_warp_marker( (Clip)self, (float)beat_time) -> None :
            Available for AudioClips only.
            Removes the specified warp marker, if possible.

            C++ signature :
                void remove_warp_marker(TPyHandle<AClip>,double)
        """

    def remove_warp_markers_listener(self, listener: Callable) -> None:
        """
        remove_warp_markers_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warp_markers".

            C++ signature :
                void remove_warp_markers_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_warp_mode_listener(self, listener: Callable) -> None:
        """
        remove_warp_mode_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warp_mode".

            C++ signature :
                void remove_warp_mode_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def remove_warping_listener(self, listener: Callable) -> None:
        """
        remove_warping_listener( (Clip)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warping".

            C++ signature :
                void remove_warping_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def replace_selected_notes(self, arg2: object) -> None:
        """
        replace_selected_notes( (Clip)arg1, (tuple)arg2) -> None :
            Called with a tuple of tuples where each inner tuple represents
            a note in the same format as returned by get_selected_notes. The
            notes described that way will then be used to replace the old selection.

            C++ signature :
                void replace_selected_notes(TPyHandle<AClip>,boost::python::tuple)
        """

    def sample_to_beat_time(self, sample_time: float) -> float:
        """
        sample_to_beat_time( (Clip)self, (float)sample_time) -> float :
            Available for AudioClips only.
            Converts the given sample time to beat time. Raises an error if the sample is not warped.

            C++ signature :
                double sample_to_beat_time(TPyHandle<AClip>,double)
        """

    def scrub(self, scrub_position: float) -> None:
        """
        scrub( (Clip)self, (float)scrub_position) -> None :
            Scrubs inside a clip.
            scrub_position defines the position in beats that the scrub will start from.
            The scrub will continue until stop_scrub is called.
            Global quantization applies to the scrub's position and length.

            C++ signature :
                void scrub(TPyHandle<AClip>,double)
        """

    def seconds_to_sample_time(self, seconds: float) -> float:
        """
        seconds_to_sample_time( (Clip)self, (float)seconds) -> float :
            Available for AudioClips only.
            Converts the given seconds to sample time. Raises an error if the sample is warped.

            C++ signature :
                double seconds_to_sample_time(TPyHandle<AClip>,double)
        """

    def select_all_notes(self) -> None:
        """
        select_all_notes( (Clip)arg1) -> None :
            Selects all notes present in the clip.

            C++ signature :
                void select_all_notes(TPyHandle<AClip>)
        """

    def select_notes_by_id(self, arg2: object) -> None:
        """
        select_notes_by_id( (Clip)arg1, (object)arg2) -> None :
            Selects all notes matching the given note IDs.

            C++ signature :
                void select_notes_by_id(TPyHandle<AClip>,boost::python::api::object)
        """

    def set_fire_button_state(self, arg2: bool) -> None:
        """
        set_fire_button_state( (Clip)arg1, (bool)arg2) -> None :
            Set the clip's fire button state directly. Supports all launch modes.

            C++ signature :
                void set_fire_button_state(TPyHandle<AClip>,bool)
        """

    def set_notes(self, arg2: object) -> None:
        """
        set_notes( (Clip)arg1, (tuple)arg2) -> None :
            Called with a tuple of tuples where each inner tuple represents
            a note in the same format as returned by get_notes. The
            notes described that way will then be added to the clip.

            C++ signature :
                void set_notes(TPyHandle<AClip>,boost::python::tuple)
        """

    def signature_denominator_has_listener(self, listener: Callable) -> bool:
        """
        signature_denominator_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "signature_denominator".

            C++ signature :
                bool signature_denominator_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def signature_numerator_has_listener(self, listener: Callable) -> bool:
        """
        signature_numerator_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "signature_numerator".

            C++ signature :
                bool signature_numerator_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def start_marker_has_listener(self, listener: Callable) -> bool:
        """
        start_marker_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "start_marker".

            C++ signature :
                bool start_marker_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def stop(self) -> None:
        """
        stop( (Clip)arg1) -> None :
            Stop playing this Clip.

            C++ signature :
                void stop(TPyHandle<AClip>)
        """

    def stop_scrub(self) -> None:
        """
        stop_scrub( (Clip)arg1) -> None :
            Stops the current scrub.

            C++ signature :
                void stop_scrub(TPyHandle<AClip>)
        """

    def velocity_amount_has_listener(self, listener: Callable) -> bool:
        """
        velocity_amount_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "velocity_amount".

            C++ signature :
                bool velocity_amount_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def warp_markers_has_listener(self, listener: Callable) -> bool:
        """
        warp_markers_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warp_markers".

            C++ signature :
                bool warp_markers_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def warp_mode_has_listener(self, listener: Callable) -> bool:
        """
        warp_mode_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warp_mode".

            C++ signature :
                bool warp_mode_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    def warping_has_listener(self, listener: Callable) -> bool:
        """
        warping_has_listener( (Clip)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warping".

            C++ signature :
                bool warping_has_listener(TPyHandle<AClip>,boost::python::api::object)
        """

    @property
    def available_warp_modes(self) -> Any:
        """
        Available for AudioClips only.
        Get/Set the available warp modes, that can be used.
        """

    @available_warp_modes.setter
    def available_warp_modes(self, value: Any) -> None:
        pass

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the Clip.
        """

    @property
    def color(self) -> Any:
        """
        Get/set access to the color of the Clip (RGB).
        """

    @color.setter
    def color(self, value: Any) -> None:
        pass

    @property
    def color_index(self) -> Any:
        """
        Get/set access to the color index of the Clip.
        """

    @color_index.setter
    def color_index(self, value: Any) -> None:
        pass

    @property
    def end_marker(self) -> Any:
        """
        Get/Set the Clips end marker pos in beats/seconds (unit depends on warping).
        """

    @end_marker.setter
    def end_marker(self, value: Any) -> None:
        pass

    @property
    def end_time(self) -> Any:
        """
        Get the clip's end time.
        """

    @property
    def file_path(self) -> Any:
        """
        Get the path of the file represented by the Audio Clip.
        """

    @property
    def gain(self) -> Any:
        """
        Available for AudioClips only.
        Read/write access to the gain setting of the
        Audio Clip
        """

    @gain.setter
    def gain(self, value: Any) -> None:
        pass

    @property
    def gain_display_string(self) -> Any:
        """
        Return a string with the gain as dB value
        """

    @property
    def groove(self) -> Any:
        """
        Get the groove associated with this clip.
        """

    @property
    def has_envelopes(self) -> bool:
        """
        Will notify if the clip gets his first envelope or the last envelope is removed.
        """

    @property
    def has_groove(self) -> bool:
        """
        Returns true if a groove is associated with this clip.
        """

    @property
    def is_arrangement_clip(self) -> bool:
        """
        return true if this Clip is an Arrangement Clip.
        A Clip can be either a Session or Arrangement Clip.
        """

    @property
    def is_audio_clip(self) -> bool:
        """
        Return true if this Clip is an Audio Clip.
        A Clip can be either an Audioclip or a MIDI Clip.
        """

    @property
    def is_midi_clip(self) -> bool:
        """
        return true if this Clip is a MIDI Clip.
        A Clip can be either an Audioclip or a MIDI Clip.
        """

    @property
    def is_overdubbing(self) -> bool:
        """
        returns true if the Clip is recording overdubs
        """

    @property
    def is_playing(self) -> bool:
        """
        Get/Set if this Clip is currently playing. If the Clips trigger mode
        is set to a quantization value, the Clip will not start playing immediately.
        If you need to know wether the Clip was triggered, use the is_triggered property.
        """

    @is_playing.setter
    def is_playing(self, value: bool) -> None:
        pass

    @property
    def is_recording(self) -> bool:
        """
        returns true if the Clip was triggered to record or is recording.
        """

    @property
    def is_triggered(self) -> bool:
        """
        returns true if the Clip was triggered or is playing.
        """

    @property
    def launch_mode(self) -> Any:
        """
        Get/Set access to the launch mode setting of the Clip.
        """

    @launch_mode.setter
    def launch_mode(self, value: Any) -> None:
        pass

    @property
    def launch_quantization(self) -> Any:
        """
        Get/Set access to the launch quantization setting of the Clip.
        """

    @launch_quantization.setter
    def launch_quantization(self, value: Any) -> None:
        pass

    @property
    def legato(self) -> Any:
        """
        Get/Set access to the legato setting of the Clip
        """

    @legato.setter
    def legato(self, value: Any) -> None:
        pass

    @property
    def length(self) -> Any:
        """
        Get to the Clips length in beats/seconds (unit depends on warping).
        """

    @property
    def loop_end(self) -> Any:
        """
        Get/Set the loop end pos of this Clip in beats/seconds (unit depends on warping).
        """

    @loop_end.setter
    def loop_end(self, value: Any) -> None:
        pass

    @property
    def loop_start(self) -> Any:
        """
        Get/Set the Clips loopstart pos in beats/seconds (unit depends on warping).
        """

    @loop_start.setter
    def loop_start(self, value: Any) -> None:
        pass

    @property
    def looping(self) -> Any:
        """
        Get/Set the Clips 'loop is enabled' flag
        .Only Warped Audio Clips or MIDI Clip can be looped.
        """

    @looping.setter
    def looping(self, value: Any) -> None:
        pass

    @property
    def muted(self) -> Any:
        """
        Read/write access to the mute state of the Clip.
        """

    @muted.setter
    def muted(self, value: Any) -> None:
        pass

    @property
    def name(self) -> str:
        """
        Read/write access to the name of the Clip.
        """

    @name.setter
    def name(self, value: str) -> None:
        pass

    @property
    def pitch_coarse(self) -> Any:
        """
        Available for AudioClips only.
        Read/write access to the pitch (in halftones) setting of the
        Audio Clip, ranging from -48 to 48
        """

    @pitch_coarse.setter
    def pitch_coarse(self, value: Any) -> None:
        pass

    @property
    def pitch_fine(self) -> Any:
        """
        Available for AudioClips only.
        Read/write access to the pitch fine setting of the
        Audio Clip, ranging from -500 to 500
        """

    @pitch_fine.setter
    def pitch_fine(self, value: Any) -> None:
        pass

    @property
    def playing_position(self) -> Any:
        """
        Constant access to the current playing position of the clip.
        The returned value is the position in beats for midi and warped audio clips,
        or in seconds for unwarped audio clips. Stopped clips will return 0.
        """

    @property
    def position(self) -> Any:
        """
        Get/Set the loop position of this Clip in beats/seconds (unit depends on warping).
        """

    @position.setter
    def position(self, value: Any) -> None:
        pass

    @property
    def ram_mode(self) -> Any:
        """
        Available for AudioClips only.
        Read/write access to the Ram mode setting of the Audio Clip
        """

    @ram_mode.setter
    def ram_mode(self, value: Any) -> None:
        pass

    @property
    def sample_length(self) -> Any:
        """
        Available for AudioClips only.
        Get the sample length in sample time or -1 if there is no sample available.
        """

    @property
    def sample_rate(self) -> Any:
        """
        Available for AudioClips only.
        Read-only access to the Clip's sampling rate.
        """

    @property
    def signature_denominator(self) -> Any:
        """
        Get/Set access to the global signature denominator of the Clip.
        """

    @signature_denominator.setter
    def signature_denominator(self, value: Any) -> None:
        pass

    @property
    def signature_numerator(self) -> Any:
        """
        Get/Set access to the global signature numerator of the Clip.
        """

    @signature_numerator.setter
    def signature_numerator(self, value: Any) -> None:
        pass

    @property
    def start_marker(self) -> Any:
        """
        Get/Set the Clips start marker pos in beats/seconds (unit depends on warping).
        """

    @start_marker.setter
    def start_marker(self, value: Any) -> None:
        pass

    @property
    def start_time(self) -> Any:
        """
        Get the clip's start time offset. For Session View clips, this is the time the clip was started. For Arrangement View clips, this is the offset within the arrangement.
        """

    @property
    def velocity_amount(self) -> Any:
        """
        Get/Set access to the velocity to volume amount of the Clip.
        """

    @velocity_amount.setter
    def velocity_amount(self, value: Any) -> None:
        pass

    @property
    def view(self) -> View:
        """
        Get the view of the Clip.
        """

    @property
    def warp_markers(self) -> Any:
        """
        Available for AudioClips only.
        Get the warp markers for this audio clip.
        """

    @property
    def warp_mode(self) -> Any:
        """
        Available for AudioClips only.
        Get/Set the warp mode for this audio clip.
        """

    @warp_mode.setter
    def warp_mode(self, value: Any) -> None:
        pass

    @property
    def warping(self) -> Any:
        """
        Available for AudioClips only.
        Get/Set if this Clip is timestreched.
        """

    @warping.setter
    def warping(self, value: Any) -> None:
        pass

    @property
    def will_record_on_start(self) -> Any:
        """
        returns true if the Clip will record on being started.
        """


class ClipLaunchQuantization:
    """
    Class that represent an enumeration of values for ClipLaunchQuantization
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    q_global = 0
    q_none = 1
    q_8_bars = 2
    q_4_bars = 3
    q_2_bars = 4
    q_bar = 5
    q_half = 6
    q_half_triplet = 7
    q_quarter = 8
    q_quarter_triplet = 9
    q_eighth = 10
    q_eighth_triplet = 11
    q_sixteenth = 12
    q_sixteenth_triplet = 13
    q_thirtysecond = 14


class GridQuantization:
    """
    Class that represent an enumeration of values for GridQuantization
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    no_grid = 0
    g_8_bars = 1
    g_4_bars = 2
    g_2_bars = 3
    g_bar = 4
    g_half = 5
    g_quarter = 6
    g_eighth = 7
    g_sixteenth = 8
    g_thirtysecond = 9
    count = 10


class LaunchMode:
    """
    Class that represent an enumeration of values for LaunchMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    trigger = 0
    gate = 1
    toggle = 2
    repeat = 3


class MidiNote:
    """
    An object representing a MIDI Note
    """

    @property
    def duration(self) -> Any:
        pass

    @property
    def mute(self) -> Any:
        pass

    @property
    def note_id(self) -> Any:
        """
        A numerical ID that's unique within the originating clip of the note. Not to be
        used directly, but important for other API calls, namely apply_note_modifications.
        """

    @property
    def pitch(self) -> Any:
        pass

    @property
    def probability(self) -> Any:
        pass

    @property
    def release_velocity(self) -> Any:
        pass

    @property
    def start_time(self) -> Any:
        pass

    @property
    def velocity(self) -> Any:
        pass

    @property
    def velocity_deviation(self) -> Any:
        pass


class MidiNoteSpecification:
    """
    An object specifying the data for creating a MIDI note. To be used with the
    add_new_notes function.
    """


class MidiNoteVector:
    """
    A container for holding MIDI notes from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (MidiNoteVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (MidiNoteVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<NClipApi::TNoteInfo, std::__1::allocator<NClipApi::TNoteInfo>> {lvalue},boost::python::api::object)
        """


class WarpMarker:
    """
    This class represents a WarpMarker type.
    """

    @property
    def beat_time(self) -> Any:
        """
        A WarpMarker's beat time.
        """

    @property
    def sample_time(self) -> Any:
        """
        A WarpMarker's sample time.
        """


class WarpMarkerVector:
    """
    A container for returning warp markers from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (WarpMarkerVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<NApiHelpers::TWarpMarker, std::__1::allocator<NApiHelpers::TWarpMarker>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (WarpMarkerVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<NApiHelpers::TWarpMarker, std::__1::allocator<NApiHelpers::TWarpMarker>> {lvalue},boost::python::api::object)
        """


class WarpMode:
    """
    Class that represent an enumeration of values for WarpMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    beats = 0
    complex = 1
    complex_pro = 2
    repitch = 3
    rex = 4
    texture = 5
    tones = 6
    count = 7



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/control.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 16119 bytes
from functools import partial

from ...base import Disconnectable, EventObject, NamedTuple, lazy_attribute, mixin, nop, old_hasattr, task

__all__ = ("Control", "InputControl", "ControlManager", "control_event", "control_color", "Connectable")


class ControlManager(EventObject):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._control_states = {}

    def add_control(self, name, control):
        if old_hasattr(self, name):
            raise AttributeError("Control would overwrite an existing property")
        control_state = control._get_state(self)
        setattr(self, name, control_state)
        return control_state

    @lazy_attribute
    def _tasks(self):
        return task.TaskGroup()

    def control_notifications_enabled(self):
        return True

    def update(self):
        for control_state in self._control_states.values():
            control_state.update()


def control_event(event_name):
    def event_decorator(self):
        def event_listener_decorator(event_listener):
            self._event_listeners[event_name] = event_listener
            return self

        return event_listener_decorator

    def event_setter(self, event_listener):
        self._event_listeners[event_name] = event_listener

    return property(event_decorator, event_setter)


class control_color:
    def __init__(self, default_color, *a, **k):
        (super().__init__)(*a, **k)
        self.default_color = default_color

    def __get__(self, obj, owner):
        if obj is None or self not in obj._colors:
            return self.default_color
        return obj._colors[self]

    def __set__(self, obj, val):
        obj._colors[self] = val
        obj._send_current_color()


class Control:
    class State(EventObject):
        enabled = True

        def __init__(self, control=None, manager=None, *a, **k):
            (super(Control.State, self).__init__)(*a, **k)
            self._colors = {}
            self._manager = manager
            self._event_listeners = control._event_listeners
            self._control_element = None
            self._has_tasks = False
            manager.register_disconnectable(self)

        def disconnect(self):
            super(Control.State, self).disconnect()
            if self._has_tasks:
                self.tasks.kill()
                self.tasks.clear()

        @lazy_attribute
        def tasks(self):
            self._has_tasks = True
            return self._manager._tasks.add(task.TaskGroup())

        @property
        def control_element(self):
            return self._control_element

        def set_control_element(self, control_element):
            self._control_element = control_element
            if self._control_element:
                self._control_element.reset_state()

        def _call_listener(self, listener_name, *args):
            listener = self._event_listeners.get(listener_name, None)
            if listener is not None:
                if self._notifications_enabled():
                    args = (*args, self)
                    listener(self._manager, *args)

        def _has_listener(self, listener_name):
            return listener_name in self._event_listeners

        def _event_listener_required(self):
            return len(self._event_listeners) > 0

        def _notifications_enabled(self):
            return self.enabled and self._manager.control_notifications_enabled()

        def update(self):
            pass

        def _send_current_color(self):
            pass

    _extra_kws = {}
    _extra_args = []

    def __init__(self, extra_args=None, extra_kws=None, *a, **k):
        (super().__init__)(*a, **k)
        self._event_listeners = {}
        if extra_args is not None:
            self._extra_args = extra_args
        if extra_kws is not None:
            self._extra_kws = extra_kws

    def __get__(self, manager, owner):
        if manager is not None:
            return self._get_state(manager)
        return self

    def __set__(self, manager, owner):
        raise RuntimeError("Cannot change control.")

    def _make_control_state(self, manager):
        return (self.State)(self._extra_args, control=self, manager=manager, **self._extra_kws)

    def _get_state(self, manager, state_factory=None):
        if self not in manager._control_states:
            if state_factory is None:
                state_factory = self._make_control_state
            manager._control_states[self] = None
            manager._control_states[self] = state_factory(manager)
        if manager._control_states[self] is None:
            raise RuntimeError("Cannot fetch state during construction of controls.")
        return manager._control_states[self]

    def _clear_state(self, manager):
        if self in manager._control_states:
            del manager._control_states[self]


class InputControl(Control):
    value = control_event("value")

    class State(Control.State):
        def __init__(self, control=None, channel=None, identifier=None, *a, **k):
            (super(InputControl.State, self).__init__)(a, control=control, **k)
            self._value_slot = None
            self._channel = channel
            self._identifier = identifier
            self._register_value_slot(self._manager, control)
            self._manager.register_disconnectable(self)

        def set_control_element(self, control_element):
            super(InputControl.State, self).set_control_element(control_element)
            if self._value_slot:
                self._value_slot.subject = control_element
            if self._control_element:
                if self._channel is not None:
                    self._control_element.set_channel(self._channel)
                if self._identifier is not None:
                    self._control_element.set_identifier(self._identifier)

        def _register_value_slot(self, manager, control):
            if self._event_listener_required():
                self._value_slot = self.register_slot(None, self._on_value, "value")

        def _on_value(self, value, *a, **k):
            self._call_listener("value", value)

        @property
        def channel(self):
            return self._channel

        @channel.setter
        def channel(self, channel):
            self._channel = channel
            if self._control_element:
                self._control_element.set_channel(self._channel)

        @property
        def identifier(self):
            return self._identifier

        @identifier.setter
        def identifier(self, value):
            self._identifier = value
            if self._control_element:
                self._control_element.set_identifier(self._identifier)


class ProxyControl:
    def __init__(self, control=None, *a, **k):
        (super().__init__)(*a, **k)
        self._control = control

    def _make_control_state(self, manager):
        return (self._control.State)(
            self._control._extra_args,
            control=self,
            manager=manager,
            **self._control._extra_kws,
        )

    def _get_state(self, manager, state_factory=None):
        return self._control._get_state(manager, self._make_control_state)

    def _clear_state(self, manager):
        self._control._clear_state(manager)


def forward_control(control):
    return mixin(ProxyControl, control.__class__)(control)


class NullSlot(Disconnectable):
    pass


class Connectable(EventObject):
    requires_listenable_connected_property = False

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._connection = self._make_empty_connection()

    def connect_property(self, subject, property_name, transform=nop):
        self.disconnect_property()
        self._connection = NamedTuple(
            slot=(self._register_property_slot(subject, property_name)),
            getter=(partial(getattr, subject, property_name)),
            setter=(partial(setattr, subject, property_name)),
            transform=transform,
        )

    def disconnect_property(self):
        self._connection.slot.disconnect()
        self._connection = self._make_empty_connection()

    def _make_empty_connection(self):
        return NamedTuple(slot=(NullSlot()), getter=nop, setter=nop, transform=nop)

    def _register_property_slot(self, subject, property_name):
        if self.requires_listenable_connected_property:
            return self.register_slot(subject, self._handle_connected_property_changed, property_name)
        return NullSlot()

    def _handle_connected_property_changed(self, value=None):
        self.on_connected_property_changed(value if value is not None else self._connection.getter())

    @property
    def connected_property_value(self):
        return self._connection.getter()

    @connected_property_value.setter
    def connected_property_value(self, value):
        self._connection.setter(self._connection.transform(value))

    def on_connected_property_changed(self, value):
        pass


class SendValueMixin:
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._value = 0

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self._value != value:
            self._value = value
            self._send_current_value()

    def set_control_element(self, control_element):
        super().set_control_element(control_element)
        self._send_current_value()

    def update(self):
        super().update()
        self._send_current_value()

    def _send_current_value(self):
        if self._control_element:
            self._control_element.send_value(self._value)


class SendValueControl(Control):
    class State(SendValueMixin, Control.State):
        pass



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/mode/selector.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 3526 bytes
from typing import Any, NamedTuple, Optional

import Live

from . import pop_last_mode


def select_mode_for_main_view(main_view_name, can_select_now=True):
    return select_mode_on_event_change(
        EventDescription(
            subject=(Live.Application.get_application().view),
            event_name="focused_document_view",
            event_state=main_view_name,
        ),
        can_select_now=can_select_now,
    )


def select_mode_on_event_change(event_description, can_select_now=False):
    subject = _get_subject(event_description.subject)
    event_name = event_description.event_name
    event_state = event_description.event_state

    def inner(modes_component, mode_name):
        def on_event_changed(*_):
            if event_state is None or getattr(subject, event_name) == event_state:
                modes_component.selected_mode = mode_name

        modes_component.register_slot(subject, on_event_changed, event_name)
        if can_select_now:
            if event_state is not None:
                on_event_changed()

    return inner


def toggle_mode_on_property_change(event_description, return_to_default=False, can_select_now=False):
    subject = _get_subject(event_description.subject)
    event_name = event_description.event_name

    def inner(modes_component, mode_name):
        def on_property_changed(state=None):
            if bool(state or getattr(subject, event_name)):
                modes_component.push_mode(mode_name)
            else:
                if return_to_default and modes_component.selected_mode == mode_name:
                    modes_component.push_mode(modes_component.modes[0])
                    modes_component.pop_unselected_modes()
                else:
                    pop_last_mode(modes_component, mode_name)

        modes_component.register_slot(subject, on_property_changed, event_name)
        if can_select_now:
            on_property_changed(getattr(subject, event_name))

    return inner


def _get_subject(subject):
    if callable(subject):
        return subject()
    return subject


class EventDescription(NamedTuple):
    subject: Any
    event_name: str
    event_state = None
    event_state: Optional[Any]



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/gcutil.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1837 bytes
import contextlib
import gc
from collections import defaultdict

from future.builtins import range

from .util import old_hasattr


def typename(obj):
    if old_hasattr(obj, "__class__"):
        return obj.__class__.__name__
    if old_hasattr(obj, "__name__"):
        return obj.__name__
    return "<unknown>"


def histogram(name_filter=None, objs=None):
    all_ = gc.get_objects() if objs is None else objs

    def _name_filter(name):
        return name_filter is None or name_filter in name

    hist = defaultdict(
        lambda: 0,
    )
    for o in all_:
        n = typename(o)
        if _name_filter(n):
            hist[n] += 1

    return hist


def instances_by_name(name_filter):
    return [o for o in gc.get_objects() if name_filter == typename(o)]


def refget(objs, level=1):
    for _ in range(level):
        refs = (gc.get_referrers)(*objs)
        with contextlib.suppress(ValueError):
            refs.remove(objs)

        objs = refs

    return refs



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/notifications/meta.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2425 bytes
from __future__ import annotations

from inspect import getmro
from typing import Callable

from .type_decl import Notification, _DefaultText, _TransformDefaultText


def transform_notification(default: Notification, transform_fn: Callable[[str], str]):
    if not callable(default):
        return transform_fn(default)

    def notification_fn(*a, **k):
        return transform_fn(default(*a, **k))

    return notification_fn


def update_special_attributes(cls, exclude_by_default=True):
    base_class = getmro(cls)[-2]
    for name, value in vars(base_class).items():
        subclass_value = vars(cls).get(name, None)
        if subclass_value is None:
            if isinstance(value, type):
                subclass_value = type(name, (value,), {})
                setattr(cls, name, subclass_value)
        if isinstance(subclass_value, type):
            include_all = getattr(subclass_value, "INCLUDE_ALL", False)
            update_special_attributes(subclass_value, False if include_all else exclude_by_default)
        if not name.startswith("__"):
            if exclude_by_default and subclass_value is None:
                setattr(cls, name, None)
            else:
                if subclass_value is None or isinstance(subclass_value, _DefaultText):
                    setattr(cls, name, getattr(base_class, name))
                if isinstance(subclass_value, _TransformDefaultText):
                    setattr(cls, name, transform_notification(getattr(base_class, name), subclass_value.transform_fn))


class DefaultNotificationsMeta(type):
    def __new__(cls, name, bases, dct):
        new_cls = super().__new__(cls, name, bases, dct)
        update_special_attributes(new_cls)
        return new_cls



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1382 bytes
from ableton.v2.control_surface.elements import ComboElement, NullFullVelocity, adjust_string

from .button import ButtonElement, SysexSendingButtonElement
from .button_matrix import ButtonMatrixElement
from .color import Color, ColorPart, ComplexColor, FallbackColor, RgbColor, SimpleColor, create_rgb_color
from .discrete_values import DiscreteValuesElement
from .display_line import DisplayLineElement
from .encoder import EncoderElement
from .lockable_button import LockableButtonElement, LockableButtonElementMixin
from .lockable_combo import LockableComboElement
from .sysex import CachingSendMessageGenerator, SysexElement
from .touch import TouchElement

__all__ = (
    "ButtonElement",
    "ButtonMatrixElement",
    "CachingSendMessageGenerator",
    "Color",
    "ColorPart",
    "ComboElement",
    "ComplexColor",
    "DiscreteValuesElement",
    "DisplayLineElement",
    "EncoderElement",
    "FallbackColor",
    "LockableButtonElement",
    "LockableButtonElementMixin",
    "LockableComboElement",
    "NullFullVelocity",
    "RgbColor",
    "SimpleColor",
    "SysexElement",
    "SysexSendingButtonElement",
    "TouchElement",
    "adjust_string",
    "create_rgb_color",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/active_parameter.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4055 bytes
from functools import partial

from ...base import find_if, listenable_property, listens_group, task
from ...live import liveobj_valid
from .. import ACTIVE_PARAMETER_TIMEOUT, Component
from ..controls import TouchControl, control_matrix
from ..display import Renderable


class ActiveParameterComponent(Component, Renderable):
    touch_controls = control_matrix(TouchControl)

    def __init__(self, timeout=ACTIVE_PARAMETER_TIMEOUT, *a, **k):
        (super().__init__)(a, name="Active_Parameter", **k)
        self._pressed_touch_elements = {}
        self._release_tasks = {}
        self._timeout = timeout

    @listenable_property
    def parameter(self):
        return find_if(
            liveobj_valid,
            (elem.controlled_parameter for elem in reversed(list(self._pressed_touch_elements.values()))),
        )

    @touch_controls.pressed
    def touch_controls(self, button):
        self._pressed_touch_elements[button.index] = button.control_element
        self._cancel_release_task(button.index)
        self._on_pressed_touch_elements_changed()

    @touch_controls.released
    def touch_controls(self, button):
        if any(t.is_pressed for t in self.touch_controls):
            self._on_touch_control_release(button.index)
        else:
            self._release_tasks[button.index] = self._tasks.add(
                task.sequence(
                    task.wait(self._timeout),
                    task.run(partial(self._on_touch_control_release, button.index)),
                ),
            )

    def _on_touch_control_release(self, index):
        self._clear_release_task(index)
        del self._pressed_touch_elements[index]
        self._on_pressed_touch_elements_changed()

    def _cancel_release_task(self, index):
        if index in self._release_tasks:
            self._release_tasks[index].kill()
            self._clear_release_task(index)

    def _on_pressed_touch_elements_changed(self):
        self._ActiveParameterComponent__on_touch_control_parameter_changed.replace_subjects(
            self._pressed_touch_elements.values(),
        )
        self._ActiveParameterComponent__on_touch_control_parameter_assignment_changed.replace_subjects(
            self._pressed_touch_elements.values(),
        )
        self.notify_parameter()

    def _clear_release_task(self, index):
        if index in self._release_tasks:
            self._tasks.remove(self._release_tasks[index])
            del self._release_tasks[index]

    @listens_group("controlled_parameter")
    def __on_touch_control_parameter_changed(self, _):
        self.notify_parameter()

    @listens_group("assignment")
    def __on_touch_control_parameter_assignment_changed(self, element):
        if not element.is_pressed:
            for index, pressed_element in self._pressed_touch_elements.items():
                if element == pressed_element:
                    if index in self._release_tasks:
                        self._cancel_release_task(index)
                        self._on_touch_control_release(index)
                    break



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4944 bytes
from base.util import CallableBool

from ableton.v2.control_surface.elements import ButtonElement as ButtonElementBase

from ...base import depends, listenable_property
from .. import MIDI_CC_TYPE
from ..display import Renderable
from ..midi import SYSEX_END


class ButtonElement(ButtonElementBase, Renderable):
    class ProxiedInterface(ButtonElementBase.ProxiedInterface):
        is_momentary = CallableBool(True)
        is_pressed = CallableBool(False)

    @depends(skin=None)
    def __init__(self, identifier, channel=0, msg_type=MIDI_CC_TYPE, is_momentary=True, led_channel=None, *a, **k):
        self._led_channel = led_channel
        (super().__init__)(is_momentary, msg_type, channel, identifier, *a, **k)
        self._do_request_rebuild = self._request_rebuild
        self._request_rebuild = self._request_rebuild_and_release

    @listenable_property
    def is_pressed(self):
        return CallableBool(self._is_momentary and int(self._last_received_value) > 0)

    @property
    def is_momentary(self):
        return CallableBool(self._is_momentary)

    def receive_value(self, value):
        was_pressed = self.is_pressed
        super().receive_value(value)
        if was_pressed != self.is_pressed:
            self.notify_is_pressed()

    def send_value(self, value, force=False, channel=None):
        channel = channel if channel is not None else self._led_channel
        super().send_value(value, force=force, channel=channel)

    def _request_rebuild_and_release(self):
        self._do_request_rebuild()
        if self.is_pressed:
            self.receive_value(0)

    def _do_draw(self, color):
        if color is not None:
            super()._do_draw(color)


class SysexSendingButtonElement(ButtonElement):
    def __init__(self, identifier, sysex_identifier, optimized=True, tail=(SYSEX_END,), *a, **k):
        (super().__init__)(identifier, *a, **k)
        self._send_message_generator = lambda *values: sysex_identifier + tuple(values) + tail
        self._optimized = optimized

    def send_value(self, *value, **_):
        message = (self._send_message_generator)(*value)
        if self._optimized:
            if message == self._last_sent_message or self.send_midi(message):
                self._last_sent_message = message
        else:
            self.send_midi(message)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3122 bytes
from .abl_signal import Signal
from .dependency import DependencyError, depends, inject
from .disconnectable import CompoundDisconnectable, Disconnectable, disconnectable
from .event import (
    Event,
    EventError,
    EventObject,
    MultiSlot,
    ObservablePropertyAlias,
    SerializableListenableProperties,
    Slot,
    SlotGroup,
    has_event,
    listenable_property,
    listens,
    listens_group,
)
from .gcutil import histogram, instances_by_name, refget
from .isclose import isclose
from .live_api_utils import (
    duplicate_clip_loop,
    is_parameter_bipolar,
    liveobj_changed,
    liveobj_valid,
    move_current_song_time,
)
from .proxy import Proxy, ProxyBase
from .util import (
    PY2,
    PY3,
    Bindable,
    BooleanContext,
    NamedTuple,
    OutermostOnlyContext,
    Slicer,
    aggregate_contexts,
    chunks,
    clamp,
    compose,
    const,
    dict_diff,
    find_if,
    first,
    flatten,
    forward_property,
    get_slice,
    group,
    in_range,
    index_if,
    infinite_context_manager,
    instance_decorator,
    is_contextmanager,
    is_iterable,
    is_matrix,
    lazy_attribute,
    linear,
    maybe,
    memoize,
    mixin,
    monkeypatch,
    monkeypatch_extend,
    negate,
    next,
    nop,
    old_hasattr,
    old_round,
    overlaymap,
    print_message,
    product,
    recursive_map,
    remove_if,
    second,
    sign,
    slice_size,
    slicer,
    third,
    to_slice,
    trace_value,
    union,
)

__all__ = (
    "Bindable",
    "BooleanContext",
    "CompoundDisconnectable",
    "DependencyError",
    "Disconnectable",
    "Event",
    "EventError",
    "EventObject",
    "MultiSlot",
    "NamedTuple",
    "ObservablePropertyAlias",
    "OutermostOnlyContext",
    "Proxy",
    "ProxyBase",
    "PY2",
    "PY3",
    "SerializableListenableProperties",
    "Signal",
    "Slicer",
    "Slot",
    "SlotGroup",
    "aggregate_contexts",
    "chunks",
    "clamp",
    "compose",
    "const",
    "depends",
    "dict_diff",
    "disconnectable",
    "duplicate_clip_loop",
    "find_if",
    "first",
    "flatten",
    "forward_property",
    "get_slice",
    "group",
    "has_event",
    "histogram",
    "in_range",
    "index_if",
    "infinite_context_manager",
    "inject",
    "instance_decorator",
    "instances_by_name",
    "is_contextmanager",
    "is_iterable",
    "is_matrix",
    "is_parameter_bipolar",
    "isclose",
    "lazy_attribute",
    "linear",
    "listenable_property",
    "listens",
    "listens_group",
    "liveobj_changed",
    "liveobj_valid",
    "maybe",
    "memoize",
    "mixin",
    "monkeypatch",
    "monkeypatch_extend",
    "move_current_song_time",
    "negate",
    "next",
    "nop",
    "overlaymap",
    "old_round",
    "old_hasattr",
    "print_message",
    "product",
    "recursive_map",
    "refget",
    "remove_if",
    "second",
    "sign",
    "slice_size",
    "slicer",
    "third",
    "to_slice",
    "trace_value",
    "union",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/recording.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 8921 bytes
from abc import ABC, abstractmethod

from Live.Song import SessionRecordStatus

from ...base import depends
from ...live import is_track_armed, is_track_recording, liveobj_valid, playing_clip_slot, prepare_new_clip_slot
from .. import Component
from ..controls import ButtonControl, ToggleButtonControl
from ..display import Renderable


class RecordingMethod(ABC):
    @depends(song=None, target_track=None)
    def __init__(self, song=None, target_track=None, *a, **k):
        (super().__init__)(*a, **k)
        self.song = song
        self.target_track = target_track

    @abstractmethod
    def trigger_recording(self):
        pass

    def start_recording(self, *_):
        self.song.session_record = True

    def stop_recording(self):
        status = self.song.session_record_status
        was_recording = status != SessionRecordStatus.off or self.song.session_record
        if was_recording:
            self.song.session_record = False
        return was_recording

    @staticmethod
    def can_record_into_clip_slot(clip_slot):
        return liveobj_valid(clip_slot) and is_track_armed(clip_slot.canonical_parent)


class BasicRecordingMethod(RecordingMethod):
    def trigger_recording(self):
        if not self.stop_recording():
            self.start_recording()


class NextSlotRecordingMethod(RecordingMethod):
    def trigger_recording(self):
        if not self.stop_recording():
            slot = prepare_new_clip_slot(self.target_track.target_track)
            if self.can_record_into_clip_slot(slot):
                slot.fire()
            else:
                self.start_recording()


class NextSlotWithOverdubRecordingMethod(NextSlotRecordingMethod):
    def trigger_recording(self):
        track = self.target_track.target_track
        playing_slot = playing_clip_slot(track)
        if not is_track_recording(track) or playing_slot is not None:
            self.song.overdub = not self.song.overdub
            self.song.is_playing = self.song.is_playing or True
        else:
            super().trigger_recording()


class RecordingComponent(Component, Renderable):
    session_record_button = ButtonControl()
    session_overdub_button = ToggleButtonControl(
        color="Recording.SessionOverdubOff",
        on_color="Recording.SessionOverdubOn",
    )
    arrangement_record_button = ToggleButtonControl(
        color="Recording.ArrangementRecordOff",
        on_color="Recording.ArrangementRecordOn",
    )
    arrangement_overdub_button = ToggleButtonControl(
        color="Recording.ArrangementOverdubOff",
        on_color="Recording.ArrangementOverdubOn",
    )
    new_button = ButtonControl(color="Recording.New", pressed_color="Recording.NewPressed")

    @depends(target_track=None)
    def __init__(self, target_track=None, recording_method_type=None, name="Recording", *a, **k):
        (super().__init__)(a, name=name, **k)
        recording_method_type = recording_method_type or BasicRecordingMethod
        self._recording_method = recording_method_type()
        song = self.song
        self.session_overdub_button.connect_property(song, "overdub")
        self.arrangement_record_button.connect_property(song, "record_mode")
        self.arrangement_overdub_button.connect_property(song, "arrangement_overdub")
        self.register_slot(song, self._update_session_record_button, "session_record_status")
        self.register_slot(song, self._update_session_record_button, "session_record")
        self._update_session_record_button()
        self._target_track = target_track
        self.register_slot(target_track, self._update_new_button, "target_clip")
        self._update_new_button()

    @session_record_button.pressed
    def session_record_button(self, _):
        self._recording_method.trigger_recording()

    @new_button.pressed
    def new_button(self, _):
        if prepare_new_clip_slot((self._target_track.target_track), stop=True):
            self.notify(self.notifications.Recording.new)

    def _update_session_record_button(self):
        song = self.song
        status = song.session_record_status
        if status == SessionRecordStatus.transition:
            self.session_record_button.color = "Recording.SessionRecordTransition"
        else:
            if status == SessionRecordStatus.on or song.session_record:
                self.session_record_button.color = "Recording.SessionRecordOn"
            else:
                self.session_record_button.color = "Recording.SessionRecordOff"

    def _update_new_button(self):
        self.new_button.enabled = liveobj_valid(self._target_track.target_clip)


class ViewBasedRecordingComponent(RecordingComponent):
    def __init__(self, name="View_Based_Recording", *a, **k):
        (super().__init__)(a, name=name, **k)
        self._record_button = None
        self._overdub_button = None
        self.register_slot(self.application.view, self.update, "focused_document_view")

    def disconnect(self):
        super().disconnect()
        self._record_button = None
        self._overdub_button = None

    def set_record_button(self, button):
        self._record_button = button
        self._update_record_button_assignments()

    def set_overdub_button(self, button):
        self._overdub_button = button
        self._update_overdub_button_assignments()

    def update(self):
        super().update()
        self._update_record_button_assignments()
        self._update_overdub_button_assignments()

    def _update_record_button_assignments(self):
        self.arrangement_record_button.set_control_element(None)
        self.session_record_button.set_control_element(None)
        if self.application.view.focused_document_view == "Session":
            self.session_record_button.set_control_element(self._record_button)
        else:
            self.arrangement_record_button.set_control_element(self._record_button)

    def _update_overdub_button_assignments(self):
        self.arrangement_overdub_button.set_control_element(None)
        self.session_overdub_button.set_control_element(None)
        if self.application.view.focused_document_view == "Session":
            self.session_overdub_button.set_control_element(self._overdub_button)
        else:
            self.arrangement_overdub_button.set_control_element(self._overdub_button)



from __future__ import annotations

from typing import Any, Callable, Optional

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class CCFeedbackRule:
    """
    Structure to define feedback properties of MIDI mappings.
    """

    @property
    def cc_no(self) -> Any:
        pass

    @property
    def cc_value_map(self) -> Any:
        pass

    @property
    def channel(self) -> Any:
        pass

    @property
    def delay_in_ms(self) -> Any:
        pass

    @property
    def enabled(self) -> Any:
        pass


class MapMode:
    """
    Class that represent an enumeration of values for MapMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    absolute = 0
    relative_signed_bit = 1
    relative_binary_offset = 2
    relative_two_compliment = 3
    relative_signed_bit2 = 4
    absolute_14_bit = 5
    relative_smooth_signed_bit = 6
    relative_smooth_binary_offset = 7
    relative_smooth_two_compliment = 8
    relative_smooth_signed_bit2 = 9


class NoteFeedbackRule:
    """
    Structure to define feedback properties of MIDI mappings.
    """

    @property
    def channel(self) -> Any:
        pass

    @property
    def delay_in_ms(self) -> Any:
        pass

    @property
    def enabled(self) -> Any:
        pass

    @property
    def note_no(self) -> Any:
        pass

    @property
    def vel_map(self) -> Any:
        pass


class PitchBendFeedbackRule:
    """
    Structure to define feedback properties of MIDI mappings.
    """

    @property
    def channel(self) -> Any:
        pass

    @property
    def delay_in_ms(self) -> Any:
        pass

    @property
    def enabled(self) -> Any:
        pass

    @property
    def value_pair_map(self) -> Any:
        pass


def forward_midi_cc(arg1: int, arg2: int, arg3: int, arg4: int, ShouldConsumeEvent: bool = True) -> bool:
    """
    forward_midi_cc( (int)arg1, (int)arg2, (int)arg3, (int)arg4 [, (bool)ShouldConsumeEvent=True]) -> bool :

        C++ signature :
            bool forward_midi_cc(unsigned int,unsigned int,int,int [,bool=True])
    """


def forward_midi_note(arg1: int, arg2: int, arg3: int, arg4: int, ShouldConsumeEvent: bool = True) -> bool:
    """
    forward_midi_note( (int)arg1, (int)arg2, (int)arg3, (int)arg4 [, (bool)ShouldConsumeEvent=True]) -> bool :

        C++ signature :
            bool forward_midi_note(unsigned int,unsigned int,int,int [,bool=True])
    """


def forward_midi_pitchbend(arg1: int, arg2: int, arg3: int) -> bool:
    """
    forward_midi_pitchbend( (int)arg1, (int)arg2, (int)arg3) -> bool :

        C++ signature :
            bool forward_midi_pitchbend(unsigned int,unsigned int,int)
    """


def map_midi_cc(
    midi_map_handle: int,
    parameter: object,
    midi_channel: int,
    controller_number: int,
    map_mode: object,
    avoid_takeover: bool,
    sensitivity: float | None = None,
) -> bool:
    """
    map_midi_cc( (int)midi_map_handle, (DeviceParameter)parameter, (int)midi_channel, (int)controller_number, (MapMode)map_mode, (bool)avoid_takeover [, (float)sensitivity=1.0]) -> bool :

        C++ signature :
            bool map_midi_cc(unsigned int,TPyHandle<ATimeableValue>,int,int,NRemoteMapperTypes::TControllerMapMode,bool [,float=1.0])
    """


def map_midi_cc_with_feedback_map(
    midi_map_handle: int,
    parameter: object,
    midi_channel: int,
    controller_number: int,
    map_mode: object,
    feedback_rule: object,
    avoid_takeover: bool,
    sensitivity: float | None = None,
) -> bool:
    """
    map_midi_cc_with_feedback_map( (int)midi_map_handle, (DeviceParameter)parameter, (int)midi_channel, (int)controller_number, (MapMode)map_mode, (CCFeedbackRule)feedback_rule, (bool)avoid_takeover [, (float)sensitivity=1.0]) -> bool :

        C++ signature :
            bool map_midi_cc_with_feedback_map(unsigned int,TPyHandle<ATimeableValue>,int,int,NRemoteMapperTypes::TControllerMapMode,NPythonMidiMap::TCCFeedbackRule,bool [,float=1.0])
    """


def map_midi_note(arg1: int, arg2: object, arg3: int, arg4: int) -> bool:
    """
    map_midi_note( (int)arg1, (DeviceParameter)arg2, (int)arg3, (int)arg4) -> bool :

        C++ signature :
            bool map_midi_note(unsigned int,TPyHandle<ATimeableValue>,int,int)
    """


def map_midi_note_with_feedback_map(arg1: int, arg2: object, arg3: int, arg4: int, arg5: object) -> bool:
    """
    map_midi_note_with_feedback_map( (int)arg1, (DeviceParameter)arg2, (int)arg3, (int)arg4, (NoteFeedbackRule)arg5) -> bool :

        C++ signature :
            bool map_midi_note_with_feedback_map(unsigned int,TPyHandle<ATimeableValue>,int,int,NPythonMidiMap::TNoteFeedbackRule)
    """


def map_midi_pitchbend(arg1: int, arg2: object, arg3: int, arg4: bool) -> bool:
    """
    map_midi_pitchbend( (int)arg1, (DeviceParameter)arg2, (int)arg3, (bool)arg4) -> bool :

        C++ signature :
            bool map_midi_pitchbend(unsigned int,TPyHandle<ATimeableValue>,int,bool)
    """


def map_midi_pitchbend_with_feedback_map(arg1: int, arg2: object, arg3: int, arg4: object, arg5: bool) -> bool:
    """
    map_midi_pitchbend_with_feedback_map( (int)arg1, (DeviceParameter)arg2, (int)arg3, (PitchBendFeedbackRule)arg4, (bool)arg5) -> bool :

        C++ signature :
            bool map_midi_pitchbend_with_feedback_map(unsigned int,TPyHandle<ATimeableValue>,int,NPythonMidiMap::TPitchBendFeedbackRule,bool)
    """


def send_feedback_for_parameter(arg1: int, arg2: object) -> None:
    """
    send_feedback_for_parameter( (int)arg1, (DeviceParameter)arg2) -> None :

        C++ signature :
            void send_feedback_for_parameter(unsigned int,TPyHandle<ATimeableValue>)
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/color.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3722 bytes
from ...base import EventObject, listenable_property, listens, liveobj_valid, nop, old_hasattr


def to_midi_value(int_or_color):
    if old_hasattr(int_or_color, "midi_value"):
        return int_or_color.midi_value
    return int(int_or_color)


class Color:
    midi_value = 0

    def __init__(self, midi_value=None, *a, **k):
        (super().__init__)(*a, **k)
        if midi_value is not None:
            self.midi_value = midi_value

    def draw(self, interface):
        interface.send_value(self.midi_value)


class DynamicColorFactory:
    def __init__(self, transformation=nop, *a, **k):
        (super().__init__)(*a, **k)
        self._transform = transformation

    def instantiate(self, song):
        raise NotImplementedError


def is_dynamic_color_factory(skin_element):
    return isinstance(skin_element, DynamicColorFactory)


class DynamicColorBase(Color, EventObject):
    midi_value = listenable_property.managed(0)

    def __init__(self, transformation=nop, *a, **k):
        (super().__init__)(*a, **k)
        self._transformation = transformation

    def _update_midi_value(self, colored_object):
        color_index = colored_object.color_index if liveobj_valid(colored_object) else None
        self.midi_value = self._transformation(color_index)


class SelectedTrackColor(DynamicColorBase):
    def __init__(self, song_view=None, *a, **k):
        (super().__init__)(*a, **k)
        self._SelectedTrackColor__on_color_changed.subject = song_view
        self._SelectedTrackColor__on_color_changed()

    @listens("selected_track.color_index")
    def __on_color_changed(self, *a):
        self._update_midi_value(self._SelectedTrackColor__on_color_changed.subject.selected_track)


class SelectedClipColor(DynamicColorBase):
    def __init__(self, song_view=None, *a, **k):
        (super().__init__)(*a, **k)
        self._SelectedClipColor__on_color_changed.subject = song_view
        self._SelectedClipColor__on_color_changed()

    @listens("detail_clip.color_index")
    def __on_color_changed(self, *a):
        view = self._SelectedClipColor__on_color_changed.subject
        self._update_midi_value(view.detail_clip)


class SelectedTrackColorFactory(DynamicColorFactory):
    def instantiate(self, song):
        return SelectedTrackColor(song_view=(song.view), transformation=(self._transform))


class SelectedClipColorFactory(DynamicColorFactory):
    def instantiate(self, song):
        return SelectedClipColor(song_view=(song.view), transformation=(self._transform))


class AnimatedColor(Color):
    _channel = None

    def __init__(self, color1=None, color2=None, *a, **k):
        (super().__init__)(*a, **k)
        self._color1 = color1
        self._color2 = color2

    def draw(self, interface):
        interface.send_value(self._color1.midi_value)
        interface.send_value((self._color2.midi_value), channel=(self._channel))


class SysexRGBColor(Color):
    def __init__(self, midi_value=None, *a, **k):
        (super().__init__)(midi_value, *a, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/isclose.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2204 bytes
import math


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    if a == b:
        return True
    if rel_tol < 0.0 or abs_tol < 0.0:
        raise ValueError("error tolerances must be non-negative")
    if math.isinf(abs(a)) or math.isinf(abs(b)):
        return False
    diff = abs(b - a)
    return diff <= abs(rel_tol * b) or diff <= abs(rel_tol * a) or diff <= abs_tol



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class DrumPad:
    """
    This class represents a drum group device pad in Live.
    """

    def add_chains_listener(self, listener: Callable) -> None:
        """
        add_chains_listener( (DrumPad)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "chains" has changed.

            C++ signature :
                void add_chains_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def add_mute_listener(self, listener: Callable) -> None:
        """
        add_mute_listener( (DrumPad)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "mute" has changed.

            C++ signature :
                void add_mute_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (DrumPad)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def add_solo_listener(self, listener: Callable) -> None:
        """
        add_solo_listener( (DrumPad)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "solo" has changed.

            C++ signature :
                void add_solo_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def chains_has_listener(self, listener: Callable) -> bool:
        """
        chains_has_listener( (DrumPad)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "chains".

            C++ signature :
                bool chains_has_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def delete_all_chains(self) -> None:
        """
        delete_all_chains( (DrumPad)arg1) -> None :
            Deletes all chains associated with a drum pad. This is equivalent to deleting a drum rack pad in Live.

            C++ signature :
                void delete_all_chains(TPyHandle<ADrumGroupDevicePad>)
        """

    def mute_has_listener(self, listener: Callable) -> bool:
        """
        mute_has_listener( (DrumPad)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "mute".

            C++ signature :
                bool mute_has_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (DrumPad)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def remove_chains_listener(self, listener: Callable) -> None:
        """
        remove_chains_listener( (DrumPad)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "chains".

            C++ signature :
                void remove_chains_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def remove_mute_listener(self, listener: Callable) -> None:
        """
        remove_mute_listener( (DrumPad)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "mute".

            C++ signature :
                void remove_mute_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (DrumPad)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def remove_solo_listener(self, listener: Callable) -> None:
        """
        remove_solo_listener( (DrumPad)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "solo".

            C++ signature :
                void remove_solo_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    def solo_has_listener(self, listener: Callable) -> bool:
        """
        solo_has_listener( (DrumPad)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "solo".

            C++ signature :
                bool solo_has_listener(TPyHandle<ADrumGroupDevicePad>,boost::python::api::object)
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the drum pad.
        """

    @property
    def chains(self) -> Any:
        """
        Return const access to the list of chains in this drum pad.
        """

    @property
    def mute(self) -> Any:
        """
        Mute/unmute the pad.
        """

    @property
    def name(self) -> str:
        """
        Return const access to the drum pad's name. It depends on the contained chains.
        """

    @property
    def note(self) -> Any:
        """
        Get the MIDI note of the drum pad.
        """

    @property
    def solo(self) -> Any:
        """
        Solo/unsolo the pad.
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ControlSurfaceComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6090 bytes
import Live

from . import Task
from .Control import ControlManager
from .Dependency import dependency, depends
from .SubjectSlot import Subject
from .Util import lazy_attribute


class ControlSurfaceComponent(ControlManager, Subject):
    name = ""
    canonical_parent = None
    is_private = False
    _show_msg_callback = dependency(show_message=None)
    _has_task_group = False
    _layer = None

    @depends(register_component=None, song=None)
    def __init__(
        self,
        name="",
        register_component=None,
        song=None,
        layer=None,
        is_enabled=True,
        is_root=False,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self.name = name
        self._explicit_is_enabled = is_enabled
        self._recursive_is_enabled = True
        self._is_enabled = self._explicit_is_enabled
        self._is_root = is_root
        self._allow_updates = True
        self._update_requests = 0
        self._song = song
        if layer is not None:
            self._layer = layer
        register_component(self)

    def disconnect(self):
        if self._has_task_group:
            self._tasks.kill()
            self._tasks.clear()
        super().disconnect()

    @property
    def is_root(self):
        return self._is_root

    def _internal_on_enabled_changed(self):
        if self._layer:
            if self.is_enabled():
                self._layer.grab(self)
            else:
                self._layer.release(self)
        if self._has_task_group:
            if self.is_enabled():
                self._tasks.resume()
            else:
                self._tasks.pause()

    def on_enabled_changed(self):
        self.update()

    def update_all(self):
        self.update()

    def set_enabled(self, enable):
        self._explicit_is_enabled = bool(enable)
        self._update_is_enabled()

    def _set_enabled_recursive(self, enable):
        self._recursive_is_enabled = bool(enable)
        self._update_is_enabled()

    def _update_is_enabled(self):
        is_enabled = self._recursive_is_enabled and self._explicit_is_enabled
        if is_enabled != self._is_enabled:
            self._is_enabled = is_enabled
            self._internal_on_enabled_changed()
            self.on_enabled_changed()

    def set_allow_update(self, allow_updates):
        allow = bool(allow_updates)
        if self._allow_updates != allow:
            self._allow_updates = allow
            if self._allow_updates:
                if self._update_requests > 0:
                    self._update_requests = 0
                    self.update()

    def control_notifications_enabled(self):
        return self.is_enabled()

    def application(self):
        return Live.Application.get_application()

    def song(self):
        return self._song

    @lazy_attribute
    @depends(parent_task_group=None)
    def _tasks(self, parent_task_group=None):
        tasks = parent_task_group.add(Task.TaskGroup())
        if not self._is_enabled:
            tasks.pause()
        self._has_task_group = True
        return tasks

    def _get_layer(self):
        return self._layer

    def _set_layer(self, new_layer):
        if self._layer != new_layer:
            if self._layer:
                self._layer.release(self)
            self._layer = new_layer
            if new_layer:
                if self.is_enabled():
                    new_layer.grab(self)

    layer = property(_get_layer, _set_layer)

    def is_enabled(self, explicit=False):
        if not explicit:
            return self._is_enabled
        return self._explicit_is_enabled

    def on_track_list_changed(self):
        pass

    def on_scene_list_changed(self):
        pass

    def on_selected_track_changed(self):
        pass

    def on_selected_scene_changed(self):
        pass

    @depends(parent_task_group=None)
    def _register_timer_callback(self, callback, parent_task_group=None):
        def wrapper(delta):
            callback()
            return Task.RUNNING

        parent_task_group.add(Task.FuncTask(wrapper, callback))

    @depends(parent_task_group=None)
    def _unregister_timer_callback(self, callback, parent_task_group=None):
        task = parent_task_group.find(callback)
        parent_task_group.remove(task)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/device_bank_registry.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1683 bytes
from ..base import EventObject, liveobj_valid


class DeviceBankRegistry(EventObject):
    __events__ = ("device_bank",)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._device_bank_registry = {}
        self._device_bank_listeners = []

    def compact_registry(self):
        newreg = dict(filter(lambda kv: liveobj_valid(kv[0]), self._device_bank_registry.items()))
        self._device_bank_registry = newreg

    def set_device_bank(self, device, bank):
        key = self._find_device_bank_key(device) or device
        old = self._device_bank_registry.get(key, 0)
        if old != bank:
            self._device_bank_registry[key] = bank
            self.notify_device_bank(device, bank)

    def get_device_bank(self, device):
        if hasattr(device, "bank_index"):
            return device.bank_index
        return self._device_bank_registry.get(self._find_device_bank_key(device), 0)

    def _find_device_bank_key(self, device):
        for k in self._device_bank_registry:
            if k == device:
                return k
        return None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/event.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 23055 bytes
import contextlib
import weakref
from functools import partial, wraps
from itertools import chain, repeat

from future.builtins import filter, zip
from future.utils import iteritems, string_types, with_metaclass

from .abl_signal import Signal
from .disconnectable import CompoundDisconnectable, Disconnectable
from .live_api_utils import liveobj_valid
from .util import NamedTuple, instance_decorator, monkeypatch, monkeypatch_extend

__all__ = (
    "EventObject",
    "Event",
    "EventError",
    "listenable_property",
    "listens",
    "listens_group",
    "Slot",
    "SlotGroup",
    "MultiSlot",
    "has_event",
    "validate_event_interface",
)


class EventError(Exception):
    pass


class Event(NamedTuple):
    name = None
    doc = ""
    signal = Signal
    override = False


def add_event(cls, event_name_or_event):
    event = Event(name=event_name_or_event) if isinstance(event_name_or_event, string_types) else event_name_or_event
    signal_attr = "_" + event.name + "_signal"

    def get_signal(self):
        try:
            return getattr(self, signal_attr)
        except AttributeError:
            signal = event.signal(sender=self)
            setattr(self, signal_attr, signal)
            return signal

    kwargs = {"doc": event.doc, "override": event.override}

    @monkeypatch(cls, (event.name + "_has_listener"), **kwargs)
    def has_method(self, slot):
        return get_signal(self).is_connected(slot)

    @monkeypatch(cls, ("add_" + event.name + "_listener"), **kwargs)
    def add_method(self, slot, identify_sender=False, *a, **k):
        sender = self if identify_sender else None
        return (get_signal(self).connect)(slot, *a, sender=sender, **k)

    @monkeypatch(cls, ("remove_" + event.name + "_listener"), **kwargs)
    def remove_method(self, slot):
        return get_signal(self).disconnect(slot)

    @monkeypatch(cls, ("notify_" + event.name), **kwargs)
    def notify_method(self, *a, **k):
        return (get_signal(self))(*a, **k)

    @monkeypatch(cls, (event.name + "_listener_count"), **kwargs)
    def listener_count_method(self):
        return get_signal(self).count

    @monkeypatch_extend(cls)
    def disconnect(self):
        get_signal(self).disconnect_all()


def validate_event_interface(obj, event_name):
    if not callable(getattr(obj, "add_" + event_name + "_listener", None)):
        raise EventError(f'Object {obj} missing "add" method for event: {event_name}')
    if not callable(getattr(obj, "remove_" + event_name + "_listener", None)):
        raise EventError(f'Object {obj} missing "remove" method for event: {event_name}')
    if not callable(getattr(obj, event_name + "_has_listener", None)):
        raise EventError(f'Object {obj} missing "has" method for event: {event_name}')


def has_event(obj, event_name):
    return (
        callable(getattr(obj, "add_" + event_name + "_listener", None))
        and callable(getattr(obj, "remove_" + event_name + "_listener", None))
        and callable(getattr(obj, event_name + "_has_listener", None))
    )


class listenable_property_base:
    def set_property_name(self, name):
        pass


class EventObjectMeta(type):
    @staticmethod
    def collect_listenable_properties(dct):
        return list(filter(lambda item: isinstance(item[1], listenable_property_base), iteritems(dct)))

    def __new__(cls, name, bases, dct):
        listenable_properties = EventObjectMeta.collect_listenable_properties(dct)
        for property_name, obj in listenable_properties:
            obj.set_property_name(property_name)

        events = dct.get("__events__", [])
        property_events = [event_name for event_name, obj in listenable_properties]
        has_events = events or property_events
        if has_events:
            if "disconnect" not in dct:
                dct["disconnect"] = lambda self: super(cls, self).disconnect()
        cls = super().__new__(cls, name, bases, dct)
        for lst in chain(events, property_events):
            add_event(cls, lst)

        return cls


class EventObject(with_metaclass(EventObjectMeta, CompoundDisconnectable)):
    def register_slot(self, *a, **k):
        if a:
            slot = a[0] if isinstance(a[0], Slot) else Slot(*a, **k)
            self.register_disconnectable(slot)
            return slot
        return None


class Slot(Disconnectable):
    _extra_kws = {}
    _extra_args = []

    def __init__(self, subject=None, listener=None, event_name=None, extra_kws=None, extra_args=None, *a, **k):
        (super().__init__)(*a, **k)
        self._event_name = event_name
        if extra_kws is not None:
            self._extra_kws = extra_kws
        if extra_args is not None:
            self._extra_args = extra_args
        self._subject = None
        self._listener = None
        self.subject = subject
        self.listener = listener

    def subject_valid(self, subject):
        return liveobj_valid(subject)

    def disconnect(self):
        self.subject = None
        self.listener = None
        super().disconnect()

    def connect(self):
        if not (self.is_connected or self.subject_valid)(self._subject) or self._listener is not None:
            add_method = getattr(self._subject, "add_" + self._event_name + "_listener")
            all_args = (*tuple(self._extra_args), self._listener)
            with contextlib.suppress(RuntimeError):
                add_method(*all_args, **self._extra_kws)

    def soft_disconnect(self):
        if not self.is_connected and self.subject_valid(self._subject) or self._listener is not None:
            all_args = (*tuple(self._extra_args), self._listener)
            remove_method = getattr(self._subject, "remove_" + self._event_name + "_listener")
            with contextlib.suppress(RuntimeError):
                remove_method(*all_args)

    @property
    def is_connected(self):
        all_args = (*tuple(self._extra_args), self._listener)
        with contextlib.suppress(RuntimeError):
            return bool(
                self.subject_valid(self._subject)
                and self._listener is not None
                and (getattr(self._subject, self._event_name + "_has_listener"))(*all_args),
            )

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, subject):
        if subject != self._subject:
            if self.subject_valid(subject):
                validate_event_interface(subject, self._event_name)
            self.soft_disconnect()
            self._subject = subject
            self.connect()

    @property
    def listener(self):
        return self._listener

    @listener.setter
    def listener(self, listener):
        if listener != self._listener:
            self.soft_disconnect()
            self._listener = listener
            self.connect()

    def __call__(self, *a, **k):
        if self._listener is not None:
            return (self._listener)(*a, **k)
        return None


class SlotGroup(EventObject):
    listener = None
    _extra_kws = None
    _extra_args = None

    def __init__(self, listener=None, event_name=None, extra_kws=None, extra_args=None, *a, **k):
        (super().__init__)(*a, **k)
        self.listener = listener
        self._event_name = event_name
        if listener is not None:
            self.listener = listener
        if extra_kws is not None:
            self._extra_kws = extra_kws
        if extra_args is not None:
            self._extra_args = extra_args

    def replace_subjects(self, subjects, identifiers=repeat(None)):
        self.disconnect()
        for subject, identifier in zip(subjects, identifiers):
            self.add_subject(subject, identifier=identifier)

    def add_subject(self, subject, identifier=None):
        if identifier is None:
            identifier = subject
        listener = self._listener_for_subject(identifier)
        self.register_slot(subject, listener, self._event_name, self._extra_kws, self._extra_args)

    def remove_subject(self, subject):
        slot = self.find_disconnectable(
            lambda x: x.subject == subject,
        )
        self.disconnect_disconnectable(slot)

    def has_subject(self, subject):
        return liveobj_valid(
            self.find_disconnectable(
                lambda x: x.subject == subject,
            ),
        )

    def _listener_for_subject(self, identifier):
        return lambda *a, **k: self.listener and (self.listener)(*(*a, identifier), **k)

    def __call__(self, *a, **k):
        return (self.listener)(*a, **k)


class MultiSlot(EventObject, Slot):
    def __init__(self, subject=None, listener=None, event_name_list=None, extra_kws=None, extra_args=None, *a, **k):
        self._original_listener = listener
        self._slot_subject = None
        self._nested_slot = None
        super().__init__(
            event_name=(event_name_list[0]),
            listener=(self._event_fired),
            subject=subject,
            extra_kws=extra_kws,
            extra_args=extra_args,
        )
        if len(event_name_list) > 1:
            self._nested_slot = self.register_disconnectable(
                MultiSlot(
                    event_name_list=(event_name_list[1:]),
                    listener=listener,
                    extra_kws=extra_kws,
                    extra_args=extra_args,
                ),
            )
            self._update_nested_subject()

    @property
    def subject(self):
        return super().subject

    @subject.setter
    def subject(self, subject):
        try:
            try:
                super(MultiSlot, MultiSlot).subject.fset(self, subject)
            except EventError:
                if self._nested_slot is None:
                    raise

        finally:
            self._slot_subject = subject
            self._update_nested_subject()

    def _event_fired(self, *a, **k):
        self._update_nested_subject()
        (self._original_listener)(*a, **k)

    def _update_nested_subject(self):
        if self._nested_slot is not None:
            self._nested_slot.subject = (
                getattr(self._slot_subject, self._event_name) if self.subject_valid(self._slot_subject) else None
            )

    def __call__(self, *a, **k):
        return (self._original_listener)(*a, **k)


def listens(event_path, *a, **k):
    @instance_decorator
    def decorator(self, method):
        event_name_list = event_path.split(".")
        if len(event_name_list) > 1:
            slot = wraps(method)(
                MultiSlot(event_name_list=event_name_list, extra_kws=k, extra_args=a, listener=(partial(method, self))),
            )
        else:
            slot = wraps(method)(
                Slot(event_name=event_path, extra_kws=k, extra_args=a, listener=(partial(method, self))),
            )
        self.register_slot(slot)
        return slot

    return decorator


def listens_group(event_name, *a, **k):
    @instance_decorator
    def decorator(self, method):
        slot = wraps(method)(
            SlotGroup(event_name=event_name, extra_kws=k, extra_args=a, listener=(partial(method, self))),
        )
        self.register_disconnectable(slot)
        return slot

    return decorator


class listenable_property(listenable_property_base, property):
    @classmethod
    def managed(cls, default_value):
        return _managed_listenable_property(default_value=default_value)


class _managed_listenable_property(listenable_property_base):
    def __init__(self, default_value=None, *a, **k):
        (super().__init__)(*a, **k)
        self._default_value = default_value
        self._property_name = None
        self._member_name = None

    def set_property_name(self, property_name):
        self._property_name = property_name
        self._member_name = f"__listenable_property_{property_name}"

    def _get_value(self, obj):
        return getattr(obj, self._member_name, self._default_value)

    def __get__(self, obj, owner):
        return self._get_value(obj)

    def __set__(self, obj, value):
        if value != self._get_value(obj):
            setattr(obj, self._member_name, value)
            getattr(obj, f"notify_{self._property_name}")(value)


class SerializableListenablePropertiesMeta(EventObjectMeta):
    def __new__(cls, name, bases, dct):
        listenable_properties = EventObjectMeta.collect_listenable_properties(dct)

        def getstate(self):
            data = super(generated_class, self).__getstate__()
            data.update({property_name: getattr(self, property_name) for property_name, _ in listenable_properties})
            return data

        def setstate(self, data):
            for k, v in iteritems(data):
                setattr(self, k, v)

        dct["__getstate__"] = getstate
        dct["__setstate__"] = setstate
        return super().__new__(cls, name, bases, dct)


class SerializableListenablePropertiesBase(Disconnectable):
    def __getstate__(self):
        return {}

    def __setstate__(self, data):
        pass


class SerializableListenableProperties(
    with_metaclass(SerializableListenablePropertiesMeta, SerializableListenablePropertiesBase),
):
    pass


class ObservablePropertyAlias(EventObject):
    def __init__(self, alias_host, property_host=None, property_name="", alias_name=None, getter=None, *a, **k):
        (super().__init__)(*a, **k)
        self._alias_host = weakref.ref(alias_host)
        self._alias_name = alias_name or property_name
        self._property_host = property_host
        self._property_name = property_name
        self._property_slot = None
        self._setup_alias(getter)

    def _get_property_host(self):
        return self._property_host

    def _set_property_host(self, host):
        self._property_host = host
        self._property_slot.subject = host

    property_host = property(_get_property_host, _set_property_host)

    def _setup_alias(self, getter):
        aliased_prop = property(getter or self._get_property)
        alias_host = self._alias_host()
        if alias_host:
            setattr(alias_host.__class__, self._alias_name, aliased_prop)
            notifier = getattr(alias_host, "notify_" + self._alias_name)
            self._property_slot = self.register_slot(Slot(self.property_host, notifier, self._property_name))

    def _get_property(self, _):
        return getattr(self.property_host, self._property_name, None)
